# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = (
    "CacheDrive",
    "CacheMemory",
)

from pathlib import Path
from typing import Union

from ..core import Bridge, Keeper
from .drive import Drive
from .memory import Memory


class CacheDrive(Bridge):
    def __init__(self, src: Keeper, root: Union[Path, str], create: bool = False):
        dst = Drive(root=root, create=create)
        super().__init__(src=src, dst=dst)


class CacheMemory(Bridge):
    def __init__(self, src: Keeper):
        dst = Memory()
        super().__init__(src=src, dst=dst)



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = ("Storage", "Keeper", "Bridge")

from typing import Any, Protocol, Type, Union, runtime_checkable

from cloudpickle import dumps  # type: ignore

from .hash import get_hash_fn
from .packaging import *


@runtime_checkable
class Keeper(Protocol):
    def exists(self, key) -> bool:
        pass

    def save(self, key, data) -> bool:
        pass

    def load(self, key) -> Any:
        pass

    def delete(self, key) -> bool:
        pass


class StorageMethod:
    def __init__(self, packaging: Type[Packaging]):
        if not issubclass(packaging, Packaging):
            raise TypeError("subclass `Packaging` expected")
        self._type = packaging

    def __get__(self, instance: "Storage", owner):
        if not issubclass(owner, Storage):
            raise ValueError
        if self.key in instance._default:
            default = instance._default[self.key]
        else:
            default = dict()
        value = self._type(**default)
        return Pipe(instance, value)

    def __set__(self, instance, value):
        raise AttributeError

    def __set_name__(self, owner, name):
        self.key = name


class Storage:
    compression = StorageMethod(Compression)
    pickling = StorageMethod(Pickling)
    encryption = StorageMethod(Encryption)
    base64 = StorageMethod(Base64)
    json = StorageMethod(Json)
    encoding = StorageMethod(Encoding)

    def __init__(
        self,
        keeper: Keeper,
        *,
        hashing: str = "blake2b",
        compression: int = -1,
        encryption_key: Union[str, bytes] = None,
        encryption_password: Union[str, bytes] = None,
        encryption_seed: int = None,
        encoding: str = "utf-8",
    ):
        if not isinstance(keeper, Keeper):
            raise ValueError
        self._keeper = keeper
        self._alg = get_hash_fn(hashing)
        self._default = dict(
            compression=dict(level=compression),
            encryption=dict(
                key=encryption_key
                or Encryption.generate_key(
                    password=encryption_password,
                    seed=encryption_seed,
                )
            ),
            encoding=dict(encoding=encoding),
        )

    def exists(self, key) -> bool:
        return self._keeper.exists(key)

    def save(self, key, data) -> bool:
        return self._keeper.save(key, data)

    def load(self, key) -> Any:
        return self._keeper.load(key)

    def delete(self, key) -> bool:
        return self._keeper.delete(key)

    def hash(self, data) -> str:
        if not isinstance(data, bytes):
            data = dumps(data)
        return self._alg(data)

    def push(self, data) -> str:
        key = self.hash(data)
        self.save(key, data)
        return key

    def pop(self, key) -> Any:
        data = self.load(key)
        self.delete(key)
        return data

    def link(self, *markers) -> "Link":
        return Link(*markers, storage=self)


class Pipe(Storage):
    def __init__(self, storage: Storage, wrapper: Packaging):
        if not isinstance(storage, Storage):
            raise ValueError
        if not isinstance(wrapper, Packaging):
            raise ValueError
        self._wrapper = wrapper
        self._default = storage._default
        self._storage = storage

    def __call__(self, **kwargs):
        self._wrapper = type(self._wrapper)(**kwargs)
        return self

    def exists(self, key) -> bool:
        return self._storage.exists(key=key)

    def save(self, key, data) -> bool:
        wrapped = self._wrapper.forward(data)
        return self._storage.save(key=key, data=wrapped)

    def load(self, key) -> Any:
        wrapped = self._storage.load(key=key)
        return self._wrapper.backward(wrapped)

    def delete(self, key) -> bool:
        return self._storage.delete(key=key)

    def hash(self, data) -> str:
        wrapped = self._wrapper.forward(data)
        return self._storage.hash(wrapped)

    def push(self, data) -> Any:
        wrapped = self._wrapper.forward(data)
        return self._storage.push(wrapped)

    def pop(self, key) -> Any:
        wrapped = self._storage.pop(key)
        return self._wrapper.backward(wrapped)


class Link:
    def __init__(self, *markers, storage: Storage):
        if not isinstance(storage, Storage):
            raise ValueError
        assert len(markers) > 0
        if len(markers) == 1:
            markers = markers[0]
        self._marker = storage.hash(markers)
        self._storage = storage

    # TODO: garbage collection

    def exists(self) -> bool:
        if not self._storage.exists(self._marker):
            return False
        data_key = self._storage.load(self._marker).decode()
        return self._storage.exists(data_key)

    def load(self) -> Any:
        data_key = self._storage.load(self._marker).decode()
        return self._storage.load(data_key)

    def delete(self) -> bool:
        data_key = self._storage.pop(self._marker).decode()
        return self._storage.delete(data_key)

    def hash(self, data) -> str:
        return self._storage.hash(data)

    def push(self, data) -> Any:
        data_key = self._storage.push(data)
        self._storage.save(self._marker, data_key.encode())
        return data_key

    def pop(self) -> Any:
        data_key = self._storage.pop(self._marker).decode()
        return self._storage.pop(data_key)


class Bridge:
    def __init__(self, src: Keeper, dst: Keeper):
        assert isinstance(src, Keeper)
        assert isinstance(dst, Keeper)
        self._src = src
        self._dst = dst if isinstance(dst, Storage) else Storage(dst)

    def exists(self, key) -> bool:
        return self._dst.link(key).exists() or self._src.exists(key)

    def save(self, key, data) -> bool:
        # check for adequacy
        saved = self._src.save(key, data)
        if not saved:
            return False
        self._dst.link(key).push(data)
        return True

    def load(self, key) -> Any:
        try:
            data = self._dst.link(key).load()
        except Exception:
            data = self._src.load(key)
            self._dst.link(key).push(data)
        return data

    def delete(self, key) -> bool:
        self._dst.link(key).delete()
        return self._src.delete(key)



from typing import Any

class Memory:
    def __init__(self) -> None: ...
    def exists(self, key) -> bool: ...
    def save(self, key, data) -> bool: ...
    def load(self, key) -> Any: ...
    def delete(self, key) -> bool: ...



import typing as tp
from pathlib import Path

class Drive:
    def __init__(self, root: tp.Union[Path, str], create: bool = ..., unstable: bool = ...) -> None: ...
    def exists(self, key: str) -> bool: ...
    def save(self, key: str, data: bytes) -> bool: ...
    def load(self, key: str) -> bytes: ...
    def delete(self, key: str) -> bool: ...

class DriveTemp(Drive):
    def __init__(self) -> None: ...
    def __del__(self) -> None: ...



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = ("Sqlite", "SqliteTemp")

from sqlitedict import SqliteDict # type: ignore
import tempfile
from pathlib import Path

class Sqlite:
    def __init__(self, path, create: bool = False):
        path = Path(path).expanduser().absolute()
        if not create and not path.exists():
            raise AssertionError
        self._db = SqliteDict(filename=path, autocommit=True)

    def __del__(self):
        self._db.close()

    def exists(self, key: str) -> bool:
        return key in self._db

    def save(self, key: str, data) -> bool:
        try:
            self._db[key] = data
            return True
        except Exception:
            return False

    def load(self, key: str):
        return self._db[key]

    def delete(self, key: str) -> bool:
        try:
            del self._db[key]
            return True
        except Exception:
            return False


class SqliteTemp(Sqlite):
    def __init__(self):
        self._temp = tempfile.NamedTemporaryFile()
        super().__init__(path=self._temp.name, create=True)

    def __del__(self):
        super().__del__()
        self._temp.close()

    def __getstate__(self):
        with open(self._temp.name, "rb") as f:
            storage = f.read()
        state = self.__dict__.copy()
        state.pop("_temp")
        state.pop("_db")
        state["__storage__"] = storage
        return state

    def __setstate__(self, state):
        temp = tempfile.NamedTemporaryFile()
        storage = state.pop("__storage__")
        with open(temp.name, "wb") as f:
            f.write(storage)

        state["_temp"] = temp
        state["_db"] = SqliteDict(filename=temp.name, autocommit=True)
        self.__dict__ = state



from typing import Any, Iterator

def iterable_hexdigest(algorithm, iterator: Iterator[bytes]) -> str: ...
def hexdigest(algorithm, data: bytes, block_size=...): ...
def get_hexdigest(algorithm): ...

registry: Any

def get_hash_fn(algorithm): ...



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = ("Memory",)

from typing import Any


class Memory:
    def __init__(self):
        self._memory = dict()

    def exists(self, key) -> bool:
        return key in self._memory

    def save(self, key, data) -> bool:
        self._memory[key] = data
        return True

    def load(self, key) -> Any:
        return self._memory[key]

    def delete(self, key) -> bool:
        if key in self._memory:
            del self._memory[key]
            return True
        return False



from ..core import Bridge, Keeper
from pathlib import Path
from typing import Union

class CacheDrive(Bridge):
    def __init__(self, src: Keeper, root: Union[Path, str], create: bool = ...) -> None: ...

class CacheMemory(Bridge):
    def __init__(self, src: Keeper) -> None: ...



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = ("bytes_to_temp_file",)

from typing import Iterator
from pathlib import Path
import tempfile
from contextlib import contextmanager


@contextmanager
def bytes_to_temp_file(data: bytes) -> Iterator[Path]:
    tempdir = tempfile.TemporaryDirectory()
    path = Path(tempdir.name) / "temp_file"
    with open(path, "wb") as file:
        file.write(data)
    try:
        yield path
    except Exception:
        pass
    finally:
        tempdir.cleanup()



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

import hashlib
import xxhash

from typing import Iterator


def iterable_hexdigest(algorithm, iterator: Iterator[bytes]) -> str:
    hasher = algorithm()
    for block in iterator:
        hasher.update(block)
    return hasher.hexdigest()


def hexdigest(algorithm, data: bytes, block_size=2 ** 20):
    if not isinstance(data, bytes):
        raise ValueError
    iterator = (data[i : i + block_size] for i in range(0, len(data), block_size))
    return iterable_hexdigest(algorithm=algorithm, iterator=iterator)


def get_hexdigest(algorithm):
    def inner(data: bytes, block_size=2 ** 20):
        return hexdigest(algorithm=algorithm, data=data, block_size=block_size)

    return inner


registry = dict(
    md5=hashlib.md5,
    sha1=hashlib.sha1,
    sha224=hashlib.sha224,
    sha256=hashlib.sha256,
    sha384=hashlib.sha384,
    sha512=hashlib.sha512,
    sha3_224=hashlib.sha3_224,
    sha3_256=hashlib.sha3_256,
    sha3_384=hashlib.sha3_384,
    sha3_512=hashlib.sha3_512,
    blake2b=hashlib.blake2b,
    blake2s=hashlib.blake2s,
    xxh32=xxhash.xxh32,
    xxh64=xxhash.xxh64,
    xxh128=xxhash.xxh128,
    xxh3_64=xxhash.xxh3_64,
    xxh3_128=xxhash.xxh3_128,
)


def get_hash_fn(algorithm):
    if isinstance(algorithm, str):
        if algorithm not in registry:
            raise ValueError
        algorithm = registry[algorithm]
    return get_hexdigest(algorithm)



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = (
    "Packaging",
    "Conveyor",
    "Compression",
    "Pickling",
    "Base64",
    "Json",
    "Encoding",
    "Encryption",
)

import base64
import os
import pickle
import zlib
import json
from hashlib import sha256
from typing import Any, Protocol, Union, runtime_checkable

import cloudpickle  # type: ignore
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


@runtime_checkable
class Packaging(Protocol):
    # TODO: checking for input and output types
    def forward(self, i) -> Any:
        pass

    def backward(self, o) -> Any:
        pass


class Conveyor:
    def __init__(self, *packer: Packaging):
        if not all([isinstance(p, Packaging) for p in packer]):
            raise TypeError
        self._packers = packer

    def forward(self, i) -> Any:
        for p in self._packers:
            i = p.forward(i)
        return i

    def backward(self, o) -> Any:
        for p in reversed(self._packers):
            o = p.backward(o)
        return o


class Compression:
    def __init__(self, level=3):
        self._level = level

    def forward(self, i: bytes) -> bytes:
        assert isinstance(i, bytes)
        return zlib.compress(i, level=self._level)

    def backward(self, o: bytes) -> bytes:
        assert isinstance(o, bytes)
        return zlib.decompress(o)


class Pickling:
    def forward(self, i) -> bytes:
        return cloudpickle.dumps(i)

    def backward(self, o: bytes) -> Any:
        assert isinstance(o, bytes)
        return pickle.loads(o)


class Base64:
    def forward(self, i: bytes) -> bytes:
        assert isinstance(i, bytes)
        return base64.urlsafe_b64encode(i)

    def backward(self, o: bytes) -> bytes:
        assert isinstance(o, bytes)
        return base64.urlsafe_b64decode(o)


class Json:
    # TODO: numpy array?
    def forward(self, i) -> str:
        return json.dumps(i)

    def backward(self, o: str) -> Any:
        assert isinstance(o, str)
        return json.loads(o)


class Encoding:
    def __init__(self, encoding="utf-8"):
        assert isinstance(encoding, str)
        self._encoding = encoding

    def forward(self, i: str) -> bytes:
        assert isinstance(i, str)
        return i.encode(encoding=self._encoding)

    def backward(self, o: bytes) -> str:
        assert isinstance(o, bytes)
        return str(o, encoding=self._encoding)


class Encryption:
    def __init__(
        self,
        *,
        key: Union[str, bytes] = None,
        password: Union[str, bytes] = None,
        seed: int = None,
    ):
        if key is not None:
            if isinstance(key, str):
                key = base64.urlsafe_b64decode(key)
            if not isinstance(key, bytes):
                raise ValueError("key must be represented in bytes")

        self._key = key or base64.urlsafe_b64decode(
            Encryption.generate_key(password=password, seed=seed)
        )

    @property
    def key(self) -> str:
        return str(base64.urlsafe_b64encode(self._key), "utf-8")

    @staticmethod
    def _hash(data: bytes) -> bytes:
        hasher = sha256()
        hasher.update(data)
        return hasher.digest()

    @staticmethod
    def _cipher(key: bytes):
        return Cipher(algorithms.AES(key), modes.ECB())

    @staticmethod
    def _padding():
        return padding.PKCS7(algorithms.AES.block_size)

    @staticmethod
    def generate_key(password: Union[str, bytes] = None, seed: int = None) -> str:
        if password is None:
            return str(base64.urlsafe_b64encode(os.urandom(32)), "utf-8")
        if seed is None:
            raise ValueError("when using a password, you must specify the salt seed")
        if isinstance(password, str):
            password = password.encode("utf-8")
        salt = Encryption._hash((seed).to_bytes(16, byteorder="big"))
        password += salt
        for _ in range(390_000):
            password = Encryption._hash(password)
        return str(base64.urlsafe_b64encode(password), "utf-8")

    def forward(self, i: bytes) -> bytes:
        assert isinstance(i, bytes)
        padder = Encryption._padding().padder()
        encryptor = Encryption._cipher(self._key).encryptor()
        padded = padder.update(i) + padder.finalize()
        return encryptor.update(padded) + encryptor.finalize()

    def backward(self, o: bytes) -> bytes:
        assert isinstance(o, bytes)
        unpadder = Encryption._padding().unpadder()
        decryptor = Encryption._cipher(self._key).decryptor()
        padded = decryptor.update(o) + decryptor.finalize()
        return unpadder.update(padded) + unpadder.finalize()



# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

__all__ = (
    "Drive",
    "DriveTemp",
)

from pathlib import Path
import typing as tp
import tempfile


class Drive:
    # TODO: split hash by dir parts
    def __init__(
        self,
        root: tp.Union[Path, str],
        create: bool = False,
        unstable: bool = True,
    ):
        root = Path(root)
        if create:
            root.mkdir(
                mode=0o750,
                parents=False,
                exist_ok=True,
            )
        assert root.exists() and root.is_dir()
        self._root = root
        self._unstable = unstable

    @staticmethod
    def _assert_type_key(key):
        assert isinstance(key, str)

    @staticmethod
    def _assert_type_data(key):
        assert isinstance(key, bytes)

    def exists(self, key: str) -> bool:
        self._assert_type_key(key)
        path = self._root / key
        if self._unstable:
            try:
                with open(path, "rb"):
                    return True
            except FileNotFoundError:
                return False
            except Exception:
                pass
        return path.exists()

    def save(self, key: str, data: bytes) -> bool:
        self._assert_type_key(key)
        self._assert_type_data(data)
        try:
            with open(self._root / key, "wb") as file:
                file.write(data)
            return True
        except Exception:
            return False

    def load(self, key: str) -> bytes:
        self._assert_type_key(key)
        with open(self._root / key, "rb") as file:
            data = file.read()
        return data

    def delete(self, key: str) -> bool:
        self._assert_type_key(key)
        try:
            Path(self._root / key).unlink()
            return True
        except Exception:
            return False


class DriveTemp(Drive):
    def __init__(self):
        self._temp = tempfile.TemporaryDirectory()
        super().__init__(
            root=self._temp.name,
            create=False,
            unstable=False,
        )

    def __del__(self):
        self._temp.cleanup()

    def __getstate__(self):
        root = self._root
        storage = dict()
        for path in root.rglob("*"):
            key = str(path.relative_to(root))
            with open(path, "rb") as file:
                storage[key] = file.read()

        state = self.__dict__.copy()
        state.pop("_temp")
        state.pop("_root")
        state["__storage__"] = storage
        return state

    def __setstate__(self, state):
        temp = tempfile.TemporaryDirectory()
        state["_temp"] = temp
        state["_root"] = Path(temp.name)
        storage = state.pop("__storage__")
        self.__dict__ = state
        for key in storage:
            self.save(key, storage[key])



"""
Redast provides a flexible interface for storing data in different types of storages.
The organization of data storage is carried out according to the principle of a data lake.
"""

# The MIT License (MIT)
# Copyright (c) 2022 Vladislav A. Proskurov
# see LICENSE for full details

from .core import *
from .local import *
from .remote import *



from typing import Any, Union

class Packaging:
    def forward(self, i) -> Any: ...
    def backward(self, o) -> Any: ...

class Conveyor:
    def __init__(self, *packer: Packaging) -> None: ...
    def forward(self, i) -> Any: ...
    def backward(self, o) -> Any: ...

class Compression:
    def __init__(self, level: int = ...) -> None: ...
    def forward(self, i: bytes) -> bytes: ...
    def backward(self, o: bytes) -> bytes: ...

class Pickling:
    def forward(self, i) -> bytes: ...
    def backward(self, o: bytes) -> Any: ...

class Base64:
    def forward(self, i: bytes) -> bytes: ...
    def backward(self, o: bytes) -> bytes: ...

class Json:
    def forward(self, i) -> str: ...
    def backward(self, o: str) -> Any: ...

class Encoding:
    def __init__(self, encoding: str = ...) -> None: ...
    def forward(self, i: str) -> bytes: ...
    def backward(self, o: bytes) -> str: ...

class Encryption:
    def __init__(self, *, key: Union[str, bytes] = ..., password: Union[str, bytes] = ..., seed: int = ...) -> None: ...
    @property
    def key(self) -> str: ...
    @staticmethod
    def generate_key(password: Union[str, bytes] = ..., seed: int = ...) -> str: ...
    def forward(self, i: bytes) -> bytes: ...
    def backward(self, o: bytes) -> bytes: ...



class Sqlite:
    def __init__(self, path, create: bool = ...) -> None: ...
    def __del__(self) -> None: ...
    def exists(self, key: str) -> bool: ...
    def save(self, key: str, data) -> bool: ...
    def load(self, key: str): ...
    def delete(self, key: str) -> bool: ...

class SqliteTemp(Sqlite):
    def __init__(self) -> None: ...
    def __del__(self) -> None: ...



"""redast installation script"""

import runpy
from pathlib import Path

from setuptools import find_packages, setup  # type: ignore

NAME = "redast"
DESCRIPTION = "remote data storage"
LICENSE = "MIT"
AUTHOR = "Vladislav A. Proskurov"
AUTHOR_EMAIL = "rilshok@pm.me"
URL = "https://github.com/rilshok/redast"
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
]

VERSION_PATH = str(Path(__file__).resolve().parent / NAME / "__version__.py")
VERSION = runpy.run_path(VERSION_PATH)["__version__"]

with open("README.md", "r", encoding="utf-8") as file:
    LONG_DESCRIPTION = file.read()

with open("requirements.txt", "r", encoding="utf-8") as file:
    REQUIREMENTS = file.read().splitlines()


setup(
    name=NAME,
    descriprion=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    version=VERSION,
    url=URL,
    packages=find_packages(include=(NAME,)),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=CLASSIFIERS,
    python_requires=">=3.8",
)



from .packaging import *
from typing import Any, Type, Union

class Keeper:
    def exists(self, key) -> bool: ...
    def save(self, key, data) -> bool: ...
    def load(self, key) -> Any: ...
    def delete(self, key) -> bool: ...

class StorageMethod:
    def __init__(self, packaging: Type[Packaging]) -> None: ...
    def __get__(self, instance: Storage, owner): ...
    def __set__(self, instance, value) -> None: ...
    key: Any
    def __set_name__(self, owner, name) -> None: ...

class Storage:
    compression: Any
    pickling: Any
    encryption: Any
    base64: Any
    json: Any
    encoding: Any
    def __init__(self, keeper: Keeper, *, hashing: str = ..., compression: int = ..., encryption_key: Union[str, bytes] = ..., encryption_password: Union[str, bytes] = ..., encryption_seed: int = ..., encoding: str = ...) -> None: ...
    def exists(self, key) -> bool: ...
    def save(self, key, data) -> bool: ...
    def load(self, key) -> Any: ...
    def delete(self, key) -> bool: ...
    def hash(self, data) -> str: ...
    def push(self, data) -> str: ...
    def pop(self, key) -> Any: ...
    def link(self, *markers) -> Link: ...

class Pipe(Storage):
    def __init__(self, storage: Storage, wrapper: Packaging) -> None: ...
    def __call__(self, **kwargs): ...
    def exists(self, key) -> bool: ...
    def save(self, key, data) -> bool: ...
    def load(self, key) -> Any: ...
    def delete(self, key) -> bool: ...
    def hash(self, data) -> str: ...
    def push(self, data) -> Any: ...
    def pop(self, key) -> Any: ...

class Link:
    def __init__(self, *markers, storage: Storage) -> None: ...
    def exists(self) -> bool: ...
    def load(self) -> Any: ...
    def delete(self) -> bool: ...
    def hash(self, data) -> str: ...
    def push(self, data) -> Any: ...
    def pop(self) -> Any: ...

class Bridge:
    def __init__(self, src: Keeper, dst: Keeper) -> None: ...
    def exists(self, key) -> bool: ...
    def save(self, key, data) -> bool: ...
    def load(self, key) -> Any: ...
    def delete(self, key) -> bool: ...
